function displayInbox() {
  folderTree.view.selection.select(0);
}

function getCurrentFolder() {
  return getFolder(folderTree.currentIndex);
}

function getFolder(row) {
  var index = 0, result;
  forEachOpenFolder(0, function(folder) {
                         if (index++ == row) {
                           result = folder;
                         }
                       });
  return result;
}

function forEachOpenFolder(parentId, onFolder, level) {
  if (!level) level = 0;
  var children = folders[parentId];
  for(var i in children) {
    var child = children[i];
    onFolder(child, level);
    if (child.open) {
      forEachOpenFolder(child.id, onFolder, level + 1);
    }
  }
}

function getFolderRow(folderId) {
  var index = 0, row;
  forEachOpenFolder(0, function(f) {
                         if (f.id == folderId) {
                           row = index;
                         }
                         index++;
                       });
  return row;
}

function getChildCount(folderId) {
  var count = 0;
  forEachOpenFolder(folderId, function() { count++; });
  return count;
}

var folderTreeView = {
  get rowCount() {
    return getChildCount(0);
  },
  getCellText: function(row, column) {
    var name = column.id.replace(/folder-/, "");
    var folder = getFolder(row);
    return SmBundle.getString(folder[name]) || folder[name];
  },
  getCellProperties: function(row, column, props) {
    var folder = getFolder(row);
    var type = (folder.id == SmFolder.TRASH) ? "trash" : "folder";
    var atom = folderTreeView.atomService.getAtom(type);
    props.AppendElement(atom);
  },
  isContainer: function(row) {
    return true;
  },
  isContainerOpen: function(row) {
    return getFolder(row).open;
  },
  isContainerEmpty: function(row) {
    var folder = getFolder(row);
    return !folders[folder.id] || folders[folder.id].length == 0;
  },
  getParentIndex: function(row) {
    var folder = getFolder(row);
    return getFolderRow(folder.parentId) || -1;
  },
  getLevel: function(row) {
    var index = 0, result;
    forEachOpenFolder(0, function(folder, level) {
                           if (index++ == row) result = level;
                         });
    return result;
  },
  hasNextSibling: function(row, afterRow) {
    var folder = getFolder(row);
    var index, result;
    forEachOpenFolder(folder.parentId,
                      function(f, level) {
                        if (f.id == folder.id) index = 0;
                        if (level == 0 &&
                            index > afterRow - row) result = true;
                        index++;
                      });
    return result;
  },
  toggleOpenState: function(row) {
    var folder = getFolder(row);
    folder.open = !folder.open;
    folderTreeView.treebox.rowCountChanged(row + 1,
      (folder.open ? 1 : -1) * getChildCount(folder.id));
  },
  DROP_ON: Components.interfaces.nsITreeView.DROP_ON,
  DROP_AFTER: Components.interfaces.nsITreeView.DROP_AFTER,

  canDrop: function(row, orientation) {
    if (orientation == folderTreeView.DROP_ON) {
      if (dragObserver.source == messageTree) return true;

      if (dragObserver.source == folderTree) {
        // Deny setting self or (sub)child as a parent
        var folder = getCurrentFolder();
        var newParent = getFolder(row);

        var result = folder.id != newParent.id;
        forEachOpenFolder(folder.id, function(f) {
                                       if (f.id == newParent.id) {
                                         result = false;
                                       }
                                     });
        return result;
      }
    }
    if (orientation == folderTreeView.DROP_AFTER &&
        dragObserver.source == folderTree &&
        row == folderTreeView.rowCount - 1) return true;
  },
  drop: function(row, orientation) {
    var target = getFolder(row);

    if (dragObserver.source == folderTree) {
      var folder = getCurrentFolder();
      var parentId = (orientation != folderTreeView.DROP_AFTER) ? target.id : 0;
      setFolderParent(folder, parentId);
    }
    else if (dragObserver.source == messageTree) {
      moveToFolder(target.id);
    }
    dragObserver.source = null;
  }
};
DefaultTreeView.extend(folderTreeView);

// Drag-n-Drop
var dragObserver = {
  source: null,
  onDragStart: function(event, transferData, action) {
    if (event.target.nodeName != "treechildren") return;

    if (event.target.parentNode == folderTree) {
      var folder = getCurrentFolder();
      if (folder.id <= SmFolder.TRASH) return false;
    }
    transferData.data = new TransferData();
    transferData.data.addDataForFlavour("text/unicode", "foo");
    dragObserver.source = event.target.parentNode;
  },
  getSupportedFlavours : function () {
    var flavours = new FlavourSet();
    flavours.appendFlavour("text/unicode");
    return flavours;
  },
  onDrop: function(event, transferData, session) {}
}

var contextFolder;

function populateFolderMenu(event) {
  var row = {}, column = {}, child = {};
  folderTreeView.treebox.getCellAt(event.clientX, event.clientY, row,
                                   column, child);
  contextFolder = (row.value != -1) ? getFolder(row.value) : null;
  var isCustomFolder = contextFolder && contextFolder.id > SmFolder.TRASH;
  $("folder-menu-rename").hidden = !isCustomFolder;
  $("folder-menu-delete").hidden = !isCustomFolder;
  $("folder-menu-empty").hidden = !contextFolder || contextFolder.id != SmFolder.TRASH;
  $("folder-menu-import").hidden = !contextFolder;
  $("folder-menu-export").hidden = !contextFolder;
}

function setFolderParent(folder, parentId) {
  if (folder.parentId == parentId) return;

  folderTreeView.treebox.beginUpdateBatch();
  if (folder.parentId != null) {
    var siblings = folders[folder.parentId];
    for(var i in siblings) {
      if (siblings[i].id == folder.id) {
        siblings.splice(i, 1);
      }
    }
    folderTreeView.treebox.rowCountChanged(folderTree.currentIndex - 1, -1);
  }

  folder.parentId = parentId;
  if (!folders[parentId]) {
    folders[parentId] = new Array();
  }
  folders[parentId].push(folder);

  var parentRow = (parentId == 0) ? parentRow = folderTreeView.rowCount - 1
                                  : getFolderRow(parentId);
  if (parentRow) {
    folderTreeView.treebox.rowCountChanged(parentRow + 1, 1);
  }
  folderTreeView.treebox.endUpdateBatch();
  storage.saveFolder(folder);
}

function addFolder() {
  var name = {};
  if (SmUtils.prompt("folderName", name) && name.value) {
    var parentId = contextFolder ? contextFolder.id : 0;
    var folder = new SmFolder(null, null, name.value);
    setFolderParent(folder, parentId);
  }
}

function renameFolder() {
  var name = { value: contextFolder.getName() };
  if (SmUtils.prompt("folderName", name) && name.value) {
    contextFolder.name = name.value;
    folderTreeView.treebox.invalidateRow(folderTree.currentIndex);
    storage.saveFolder(contextFolder);
  }
}

function deleteFolder() {
  if (contextFolder.parentId != SmFolder.TRASH &&
      SmPrefs.getBoolPref("moveToTrash"))
  {
    return setFolderParent(contextFolder, SmFolder.TRASH);
  }
  var name = contextFolder.getName();
  var confirmText = SmBundle.getString("deleteFolder", [name]);
  askDeleteFolder(contextFolder, confirmText);
}

function emptyTrash() {
  var confirmText = SmBundle.getString("emptyTrash");
  askDeleteFolder(contextFolder, confirmText, true);
}

function askDeleteFolder(folder, confirmText, childrenOnly) {

  function forEachFolder(folder, onFolder) {
    var children = folders[folder.id];
    for(var i in children) {
      forEachFolder(children[i], onFolder);
    }
    onFolder(folder);
  }

  function doDeleteFolder() {
    storage.deleteFolder(folder, childrenOnly);

    // Update tree: delete child folders
    var folderRow = getFolderRow(folder.id);
    var children = folderRow && folder.open ? getChildCount(folder.id) : 0;
    forEachFolder(folder, function(f) {
      delete folders[f.id];
    });
    folderTreeView.treebox.rowCountChanged(folderRow, -children);

    if (childrenOnly) {
      if (folderRow == folderTree.currentIndex) displayFolder();
      return;
    }

    // Update tree: delete folder itself
    var siblings = folders[folder.parentId];
    for(var i in siblings) {
      if (siblings[i].id == folder.id) {
        siblings.splice(i, 1);
        folderTreeView.treebox.rowCountChanged(folderRow - 1, -1);
      }
    }
  }

  function getFolderMessages() {
    var msgs = new Array();
    forEachFolder(folder, function(f) {
      msgs = msgs.concat(storage.getMessages(f.id));
    });
    return msgs;
  }

  var msgs = getFolderMessages();
  if (msgs.length) {
    SimpleMail.askDeleteMail(msgs, confirmText, window, function() {
      if (getFolderMessages().length == 0) doDeleteFolder();
    });
  }
  else if (SmUtils.confirm(confirmText)) doDeleteFolder();
}

function getFolderName(id) {
  for(var i in folders) {
    var children = folders[i];
    for (var j in children) {
      if (children[j].id == id) return children[j].getName();
    }
  }
}

function exportFolder() {
  var outFile = SmFile.chooseFile("save");
  if (!outFile) return;

  var out = SmFile.createOutputStream(outFile);

  var msgs = storage.getMessages(getCurrentFolder().id);
  for(var i = 0; i < msgs.length; i++) {
    var file = SmFile.getAttachmentsDirectory(msgs[i].attachmentsDir);
    file.append("message.txt");

    out.write("From - " + new Date(msgs[i].date) + "\r\n");
    out.write(SmFile.readFile(file) + "\r\n\r\n");
  }
  out.close();
}

function importFolder() {
  var inFile = SmFile.chooseFile("open");
  if (!inFile) return;

  var folderId = getCurrentFolder().id;
  var contents = SmFile.readFile(inFile);
  var chunks = contents.split(/From .*\r\n/);

  for(var i = 1; i < chunks.length; i++) {
    var chunk = chunks[i].replace(/\r\n\r\n$/, "");
    var message = SmMessage.parse(chunk);
    message.folderId = folderId;
    storage.saveMessage(message);
  }
  displayFolder();
}