var messageTreeView = {
  get rowCount() {
    return messages.length;
  },

  getCellText: function(row, column) {
    if (column.id == "image" ||
        column.id == "attachmentsCount") return null;

    var value = messages[row][column.id];
    switch(column.id) {
           case "date": return new Date(value).toLocaleString();
           case "size": return SmUtils.size2string(value);
       case "folderId": return getFolderName(value);
          case "color": return "          ";
               default: return value;
    }
  },

  getImageSrc: function(row, column) {
    if (column.id != "image") return null;

    var message = messages[row];
    var account = storage.getAccount(message.accountId);
    return !account ? SmFile.getStatusImageURL(message.read ? "noMail" : "mail")
                    : SmFile.getImageURL(message.read ? account.getReadIcon()
                                                      : account.getUnreadIcon());
  },

  getCellProperties: function(row, column, props) {
    var message = messages[row];
    if (!message.read) {
      var atom = messageTreeView.atomService.getAtom("unread");
      props.AppendElement(atom);
    }
    if (column.id == "attachmentsCount" && messages[row].attachmentsCount) {
      var atom = messageTreeView.atomService.getAtom("attachment");
      props.AppendElement(atom);
    }
    if (column.id == "color") {
      var atom = messageTreeView.atomService.getAtom("color");
      props.AppendElement(atom);
    }
    if (message.color) {
      var colors = ["", "red", "orange", "green", "blue", "violet"];
      var atom = messageTreeView.atomService.getAtom(colors[message.color]);
      props.AppendElement(atom);
    }
  },

  getRowProperties: function(row, props) {
    var message = messages[row];
    if (message.color) {
      var colors = ["", "red", "orange", "green", "blue", "violet"];
      var atom = messageTreeView.atomService.getAtom(colors[message.color]);
      props.AppendElement(atom);
    }
  },

  cycleHeader: function(column) {
    column = $(column.id);
    var sortedColumn = messageTreeView.getSortedColumn();
    if (sortedColumn && sortedColumn != column) {
      sortedColumn.removeAttribute("sortDirection");
    }            
    var direction = column.getAttribute("sortDirection");
    column.setAttribute("sortDirection",
                        messageTreeView.nextDirection[direction]);
    displayFolder();
  },

  action: function() {
    if (isPartialSelected()) downloadMessageBody();
    else if (getCurrentFolder().id != SmFolder.DRAFTS) toggleRead();
    else {
      var message = storage.getMessage(messages[messageTree.currentIndex].id);
      SimpleMail.openNewMessage(message);
    }
  },
  delete: function() { deleteMail(); },
};
DefaultTreeView.extend(messageTreeView);

function displayMessageHtml() {
  var body = $("body");
  body.contentWindow.scroll(0, 0);

  if (messageTree.currentIndex == -1) {
    body.contentDocument.body.innerHTML = "";
    return;
  }
  var message = messages[messageTree.currentIndex];
  var html = message.partial ? SmBundle.getString("messageNotDownloaded")
                             : storage.getMessage(message.id).html;
  var search = $("search").value;
  if (search) html = SmUtils.highlight(search, html);
  body.contentDocument.body.innerHTML = html;

  if (message.returnReceiptTo &&
      message.uid && !message.partial) sendReturnReceipt(message);

  startMarkReadTimeout(message);
}

function sendReturnReceipt(message) {
  function clearReturnReceiptTo() {
    message.returnReceiptTo = null;
    storage.saveMessageState(message);
  }
  var prompts = Components.classes["@mozilla.org/embedcomp/prompt-service;1"]
                .getService(Components.interfaces.nsIPromptService);
  var button = prompts.confirmEx(window, "Simple Mail",
                                 SmBundle.getString("sendReturnReceipt"),
                                 prompts.BUTTON_POS_0 * prompts.BUTTON_TITLE_YES +
                                 prompts.BUTTON_POS_2 * prompts.BUTTON_TITLE_NO +
                                 prompts.BUTTON_POS_1 * prompts.BUTTON_TITLE_IS_STRING,
                                 null, SmBundle.getString("later"), null, null, {});
  if (button == 0) { // "Yes"
    var account = storage.getAccount(message.accountId);
    var receipt = message.createReturnReceipt(account.toString());
    SimpleMail.sendMail(receipt, clearReturnReceiptTo);
  }
  else if (button == 2) { // "No"
    clearReturnReceiptTo();
  }
}

var markReadTimeout;

// Mark message read after 10 seconds of inactivity
function startMarkReadTimeout(message) {
  clearTimeout(markReadTimeout);
  if (!message.read && !message.partial) {
    markReadTimeout = setTimeout(function() {
      if (messages[messageTree.currentIndex] &&
          messages[messageTree.currentIndex].id == message.id)
      {
        markRead(messageTree.currentIndex, true);
      }
    }, 10*1000);
  }
}

// Mark read individual message
function markRead(index, read) {
  var message = messages[index];
  message.read = read;
  storage.saveMessageState(message);
  messageTreeView.treebox.invalidateRow(index);
}

function displayMessageSource() {
  var attachmentsDir = messages[messageTree.currentIndex].attachmentsDir;
  var gBrowser = SmUtils.getThisWindow().browser;
  gBrowser.selectedTab = gBrowser.addTab(SmFile.getAttachmentURL(attachmentsDir, "message.txt"));
}

// If there's an unread message selected, mark all selected messages read.
// Otherwise mark all selected messages unread.

function toggleRead() {
  var read = false;
  messageTreeView.forEachSelectedItem(function(index) {
    read |= !messages[index].read;
  });
  messageTreeView.forEachSelectedItem(function(index) {
    markRead(index, read);
  });
}

function deleteMail() {
  if (messageTree.view.selection.getRangeCount() == 0) return;

  if (getCurrentFolder().id != SmFolder.TRASH &&
      SmPrefs.getBoolPref("moveToTrash")) return moveToFolder(SmFolder.TRASH);

  var msgs = new Array();
  messageTreeView.forEachSelectedItem(function(index) {
    msgs.push(messages[index]);
  });
  var confirmText = SmBundle.getString("deleteMessages");
  SimpleMail.askDeleteMail(msgs, confirmText, window);
}

function onMessagesDeleted(deleted) {
  if (!deleted) return;

  function isDeleted(message) {
    for(var i in deleted) {
      if (deleted[i].id == message.id) return true;
    }
  }
  messageTreeView.treebox.beginUpdateBatch();
  if (messageTree.currentIndex != -1 && 
      isDeleted(messages[messageTree.currentIndex]))
  {
    messageTree.currentIndex = -1; // Otherwise "onselect" fires during delete
  }
  for(var current = 0, i = messages.length - 1; i >= 0; i--) {
    if (isDeleted(messages[i])) {
      messages.splice(i, 1);
      messageTreeView.treebox.rowCountChanged(i, -1);
      if (messageTree.currentIndex == -1 && i < messages.length) {
        messageTree.view.selection.select(i);
      }
    }
  }
  if (messageTree.currentIndex == -1 && messages.length > 0) {
    messageTree.view.selection.select(messages.length - 1);
  }
  messageTreeView.treebox.endUpdateBatch();
}

function onMessageAdded(message) {
  if (message.folderId == getCurrentFolder().id) {
    messages.splice(0, 0, message);
    messageTreeView.treebox.rowCountChanged(0, 1);
  }                              
}

function onMessageUpdated(message) {
  for (var i = 0; i < messages.length; i++) {
    if (messages[i].id == message.id) {
      messages[i] = message;
      messageTreeView.treebox.invalidateRow(i);
      if (i == messageTree.currentIndex) displayMessageHtml();
      break;
    }
  }
}

function populateRefreshMenu() {
  var accounts = storage.getAccounts();
  if (!accounts.length) return false;

  var refresh = $("refresh");
  SmUtils.clearElement(refresh);

  for(var i in accounts) {
    var account = accounts[i];
    if (!account.pop3.host) continue;

    var menuitem = document.createElement("menuitem");
    menuitem.setAttribute("label", account);
    menuitem.setAttribute("class", "menuitem-iconic");
    menuitem.setAttribute("image", SmFile.getImageURL(account.getUnreadIcon()));
    menuitem.setAttribute("oncommand", "SimpleMail.getMail(" + account.id + ");");
    refresh.appendChild(menuitem);
  }
}

function populateRecentMenu(menupopup, method) {
  var recent = SimpleMail.storage.getRecentAddresses();
  if (!recent.length) return false;

  SmUtils.clearElement(menupopup);

  for(var i in recent) {
    var menuitem = document.createElement("menuitem");
    menuitem.setAttribute("label", recent[i]);
    menuitem.setAttribute("oncommand", method + "('" + recent[i] + "');");
    menupopup.appendChild(menuitem);
  }
}

function openReply(all) {
  if (messageTree.currentIndex != -1) {
    var message = messages[messageTree.currentIndex];
    var reply = storage.getMessage(message.id).createReply(all);
    SimpleMail.openNewMessage(reply);
  }
}

function openNew(to) {
  var message = new SmMessage();
  message.to = to;
  SimpleMail.openNewMessage(message);
}

function openForward(to) {
  if (messageTree.currentIndex != -1) {
    var message = messages[messageTree.currentIndex];
    var forward = storage.getMessage(message.id).createForward();
    forward.to = to;
    SimpleMail.openNewMessage(forward);
  }
}

function displayFolder() {
  var search = $("search").value;
  $("folderId").hidden = !search;

  var oldCount = messages.length;
  var folder = getCurrentFolder();
  messages = storage.getMessages(folder && folder.id, search, getSort());
  messageTreeView.treebox.rowCountChanged(0, messages.length - oldCount);
  messageTreeView.treebox.invalidate();
  if (messageTreeView.rowCount > 0) {
    messageTree.view.selection.select(0);
    messageTreeView.treebox.scrollToRow(0);
    // onSelect event didn't fire, need to call displayMessageHtml() directly
    if (messages.length - oldCount == 0) displayMessageHtml();
  }
}

function moveToFolder(folderId) {
  messageTreeView.forEachSelectedItem(function(index) {
    messages[index].folderId = folderId;
    storage.saveMessageState(messages[index]);
    messages.splice(index, 1);
    messageTreeView.treebox.rowCountChanged(index, -1);
  });
}

function getSort() {
  var column = messageTreeView.getSortedColumn(); 
  if (column) {
    var direction = column.getAttribute("sortDirection");
    if (direction) {
      var sortColumns = { "image": "accountId",
                          "from" : "fromm",
                          "to"   : "too",
                          "color": "noColor, color" };

      return (sortColumns[column.id] || column.id) + " " +
             (direction == "descending" ? "DESC" : "ASC");
    }
  }
}

// Partially downloaded messages

function populateMessageMenu() {
  $("message-menu-download").hidden = !isPartialSelected();
  return messageTree.view.selection.getRangeCount() > 0;
}

function isPartialSelected() {
  var isPartial = false;
  messageTreeView.forEachSelectedItem(function(index) {
    if (messages[index].partial) isPartial = true;
  });
  return isPartial;
}

function downloadMessageBody() {
  var partial = new Array();
  messageTreeView.forEachSelectedItem(function(index) {
    var message = messages[index];
    if (message.partial) partial[message.id] = message;
  });
  SimpleMail.getMailBody(partial);
}

function setMessageColor(color) {
  messageTreeView.forEachSelectedItem(function(index) {
    var message = messages[index];
    message.color = color;
    storage.saveMessageState(message);
    messageTreeView.treebox.invalidateRow(index);
  });
}

function printMessage() {
  if (messageTree.currentIndex != -1) {
    $("body").contentWindow.print();
  }
}
