function SmPasswordManagerImpl() { // Firefox 2.0
  var passwordManager = Components.classes["@mozilla.org/passwordmanager;1"]
                        .getService(Components.interfaces.nsIPasswordManager);
  var passwordManagerInternal = Components.classes["@mozilla.org/passwordmanager;1"]
                                .getService(Components.interfaces.nsIPasswordManagerInternal);

  var url = "chrome://simplemail/";
  var userName = "simplemail";

  this.getPassword = function(id) {
    var host = { value: url + id };
    var user = { value: userName };
    var password = { value: "" }; 
    try {
      passwordManagerInternal.findPasswordEntry(host.value, user.value,
                                                password.value, host, user, password);
    }
    catch(e) {}
    return password.value;
  }

  this.removePassword = function(id) {
    try {
      passwordManager.removeUser(url + id, userName);
    }
    catch(e) {}
  }

  this.setPassword = function(id, password) {
    this.removePassword(id);
    try {
      passwordManager.addUser(url + id, userName, password);
    }
    catch(e) {}
  }
}

function SmLoginManagerImpl() { // Firefox 3.0
  var loginManager = Components.classes["@mozilla.org/login-manager;1"]
                     .getService(Components.interfaces.nsILoginManager);

  var nsLoginInfo = new Components.Constructor("@mozilla.org/login-manager/loginInfo;1",
                                               Components.interfaces.nsILoginInfo,
                                               "init");
  var url = "chrome://simplemail/";
  var realm = "Simple Mail Account Password";
  var userName = "simplemail";

  this.getPassword = function(id) {
    var logins = loginManager.findLogins({}, url + id, realm, null);
      
    for (var i = 0; i < logins.length; i++) {
      if (logins[i].username == userName) {
         return logins[i].password;
      }
    }
    return "";
  }

  this.removePassword = function(id) {
    var logins = loginManager.findLogins({}, url + id, realm, null);
      
    for (var i = 0; i < logins.length; i++) {
      if (logins[i].username == userName) {
         loginManager.removeLogin(logins[i]);
         return;
      }
    }
  }

  this.setPassword = function(id, password) {
    this.removePassword(id);
    try {
      var loginInfo = new nsLoginInfo(url + id, realm, null,
                                      userName, password, null, null);
      loginManager.addLogin(loginInfo);
    }
    catch(e) {}
  }
}

var SmPasswordManager = Components.classes["@mozilla.org/passwordmanager;1"]
                        ? new SmPasswordManagerImpl()
                        : new SmLoginManagerImpl();



function SmStorage() {
  var file = SmFile.getDataDirectory();
  file.append("simplemail.sqlite");
  this.firstRun = !file.exists();

  var storageService = Components.classes["@mozilla.org/storage/service;1"]
                       .getService(Components.interfaces.mozIStorageService);
  var connection = storageService.openDatabase(file);

  if (this.firstRun) {
    connection.executeSimpleSQL("CREATE TABLE addressBook (" +
                                  "id               INTEGER PRIMARY KEY AUTOINCREMENT," +
                                  "name             STRING," +
                                  "email            STRING," +
                                  "sentDate         DATETIME)");

    connection.executeSimpleSQL("CREATE INDEX addressBook_email ON addressBook (email)");
    connection.executeSimpleSQL("CREATE INDEX addressBook_sentDate ON addressBook (sentDate)");

    connection.executeSimpleSQL("CREATE TABLE accounts (" + 
                                  "id               INTEGER PRIMARY KEY AUTOINCREMENT," +
                                  "name             STRING," +
                                  "email            STRING," +
                                  "refresh          DOUBLE," +
                                  "deleteMode       STRING," +
                                  "color            STRING," +
                                  "signature        STRING," +
                                  "isImap           BOOLEAN," +
                                  "pop3auth         INTEGER," +
                                  "pop3host         STRING," +
                                  "pop3port         INTEGER," +
                                  "pop3ssl          BOOLEAN," +
                                  "pop3login        STRING," +
                                  "smtpHost         STRING," +
                                  "smtpPort         INTEGER," +
                                  "smtpSsl          BOOLEAN," +
                                  "smtpLogin        STRING)");

    connection.executeSimpleSQL("CREATE TABLE messages (" + 
                                  "id               INTEGER PRIMARY KEY AUTOINCREMENT," +
                                  "accountId        INTEGER," +
                                  "uid              STRING," +
                                  "fromm            STRING," + // "from" is reserved word
                                  "too              STRING," + // "to" is reserved word
                                  "cc               STRING," + 
                                  "bcc              STRING," + 
                                  "subject          STRING," + 
                                  "date             DATETIME," + 
                                  "size             DOUBLE," + 
                                  "charset          STRING," +
                                  "read             BOOLEAN," +
                                  "deleted          BOOLEAN," +
                                  "attachmentsDir   STRING," +
                                  "attachmentsCount INTEGER," +
                                  "html             STRING," +
                                  "folderId         INTEGER," +
                                  "partial          BOOLEAN," +
                                  "template         BOOLEAN," +
                                  "color            INTEGER DEFAULT 0," +
                                  "returnReceiptTo  STRING)");

    connection.executeSimpleSQL("CREATE INDEX messages_accountId_uid ON messages (accountId, uid)");
    connection.executeSimpleSQL("CREATE INDEX messages_deleted_folderId_id ON messages (deleted, folderId, id DESC)");

    connection.executeSimpleSQL("CREATE TABLE folders (" + 
                                  "id               INTEGER PRIMARY KEY AUTOINCREMENT," +
                                  "parentId         INTEGER DEFAULT 0," +
                                  "name             STRING)");
    // Create default folders
    var statement = connection.createStatement("INSERT INTO folders (name) VALUES (?1)");
    for(var i in SmFolder.DEFAULT_NAMES) {
      statement.bindStringParameter(0, SmFolder.DEFAULT_NAMES[i]);
      statement.execute();
    }
  }

  try { // Since version 2.00
    connection.executeSimpleSQL("ALTER TABLE messages ADD bcc STRING");
    connection.executeSimpleSQL("ALTER TABLE messages ADD folderId INTEGER DEFAULT " + SmFolder.INBOX); // Set Inbox as a folder for all legacy messages
    connection.executeSimpleSQL("ALTER TABLE messages ADD partial BOOLEAN");
    connection.executeSimpleSQL("ALTER TABLE messages ADD template BOOLEAN");
    connection.executeSimpleSQL("DROP INDEX messages_deleted_id");
    connection.executeSimpleSQL("CREATE INDEX messages_deleted_folderId_id ON messages (deleted, folderId, id DESC)");
  }
  catch(e) {}

  try { // Since version 2.02
    connection.executeSimpleSQL("ALTER TABLE addressBook ADD sentDate DATETIME");
  }
  catch(e) {}

  try { // Since version 2.03
    connection.executeSimpleSQL("ALTER TABLE accounts ADD color STRING");
  }
  catch(e) {}

  try { // Since version 2.13
    connection.executeSimpleSQL("CREATE TABLE folders (" + 
                                  "id       INTEGER PRIMARY KEY AUTOINCREMENT," +
                                  "parentId INTEGER DEFAULT 0," +
                                  "name     STRING)");
    // Create default folders
    var statement = connection.createStatement("INSERT INTO folders (name) VALUES (?1)");
    for(var i in SmFolder.DEFAULT_NAMES) {
      statement.bindStringParameter(0, SmFolder.DEFAULT_NAMES[i]);
      statement.execute();
    }
  }
  catch(e) {}

  try { // Since version 2.31
    connection.executeSimpleSQL("ALTER TABLE accounts ADD isImap BOOLEAN");
  }
  catch(e) {}

  try { // Since version 2.35
    connection.executeSimpleSQL("ALTER TABLE messages ADD attachmentsCount INTEGER");
  }
  catch(e) {}

  try { // Since version 2.37
    connection.executeSimpleSQL("ALTER TABLE messages ADD color INTEGER default 0");
  }
  catch(e) {}

  try { // Since version 2.39
    connection.executeSimpleSQL("ALTER TABLE accounts ADD pop3auth INTEGER");
  }
  catch(e) {}

  try { // Since version 2.43
    connection.executeSimpleSQL("ALTER TABLE messages ADD returnReceiptTo STRING");
  }
  catch(e) {}

  var text = "text: ";

  function bindString(statement, id, value) {
    // Force 'text' type for strings with leading zeroes, otherwise SQLite strips these zeroes
    statement.bindStringParameter(id, (value && value[0] == "0") ? text + value : value);
  }

  function getString(statement, id) {
    var value = statement.getString(id);
    return (value && value.substr(0, text.length) == text) ? value.substr(text.length) : value;
  }

  function getAddress(email) {
    var sql = "SELECT id, name, email, sentDate FROM addressBook WHERE email=?1";
    var statement = connection.createStatement(sql);
    try {
      bindString(statement, 0, email);
      return statement.executeStep() ? new SmAddress(statement.getInt32(0),
                                                     getString(statement, 1),
                                                     getString(statement, 2),
                                                     statement.getInt64(3))
                                     : null;
    }
    finally {
      statement.reset();
    }
  }

  this.saveAddress = function(address) {
    // If address already exists, update it
    var exists = getAddress(address.email);
    if (exists) address.id = exists.id;

    var sql = address.id ? "UPDATE addressBook SET name=?1, email=?2, sentDate=?3 WHERE id=?4"
                         : "INSERT INTO addressBook (name, email, sentDate) VALUES (?1, ?2, ?3)";

    var statement = connection.createStatement(sql);

    bindString(statement, 0, address.name);
    bindString(statement, 1, address.email);
    statement.bindInt64Parameter(2, new Date().getTime());
    if (address.id) statement.bindInt32Parameter(3, address.id);
    statement.execute();

    if (!address.id) {
      var statement = connection.createStatement("SELECT last_insert_rowid() FROM addressBook");
      try {
        if (statement.executeStep()) address.id = statement.getInt32(0);
      }
      finally {
        statement.reset();
      }
    }
  }

  this.addAddressList = function(list) {
    var addresses = SmAddress.parseList(list);
    for(var i in addresses) {
      this.saveAddress(addresses[i]);
    }
  }

  function getAddresses(sql) {
    var statement = connection.createStatement(sql);
    try {
      var addresses = new Array();
      while(statement.executeStep()) {
        addresses.push(new SmAddress(statement.getInt32(0),
                                     getString(statement, 1),
                                     getString(statement, 2),
                                     statement.getInt64(3)));
      }
      return addresses;
    }
    finally {
      statement.reset();
    }
  }

  this.getAddresses = function() {
    return getAddresses("SELECT id, name, email, sentDate FROM addressBook ORDER BY name, email");
  }                               

  this.getRecentAddresses = function() {
    return getAddresses("SELECT id, name, email, sentDate FROM addressBook ORDER BY sentDate DESC LIMIT 10");
  }

  this.deleteAddress = function(address) {
    var statement = connection.createStatement("DELETE FROM addressBook WHERE id=?1");
    statement.bindInt32Parameter(0, address.id);
    statement.execute();
  }

  var accounts; // Accounts list cache

  this.getAccounts = function() {
    if (!accounts) {
      accounts = doGetAccounts();
    }
    return accounts;
  }

  this.getAccount = function(id) {
    var accounts = this.getAccounts();
    for(var i in accounts) {
      if (accounts[i].id == id) return accounts[i];
    }
  }

  function doGetAccounts() {
    var accounts = new Array();
    var sql = "SELECT id, name, email, refresh, deleteMode, color, signature, isImap, pop3auth," +
              "pop3host, pop3port, pop3ssl, pop3login, " +
              "smtpHost, smtpPort, smtpSsl, smtpLogin FROM accounts";
    var statement = connection.createStatement(sql);
    try {
      while(statement.executeStep()) {
        var id = statement.getInt32(0);
        accounts.push(new SmAccount(id,
                                    getString(statement, 1),
                                    getString(statement, 2),
                                    statement.getDouble(3),
                                    getString(statement, 4),
                                    getString(statement, 5),
                                    getString(statement, 6),
                                    statement.getInt32(7),
                                    statement.getInt32(8),
                                    getString(statement, 9),
                                    statement.getInt32(10),
                                    statement.getInt32(11),
                                    getString(statement, 12),
                                    getString(statement, 13),
                                    statement.getInt32(14),
                                    statement.getInt32(15),
                                    getString(statement, 16)));
      }
    }
    finally {
      statement.reset();
    }
    return accounts;
  }

  this.saveAccount = function(account) {
    var sql = account.id ? "UPDATE accounts SET name=?1, email=?2, refresh=?3, deleteMode=?4," +
                           "color=?5, signature=?6, isImap=?7, pop3auth=?8," + 
                           "pop3host=?9, pop3port=?10, pop3ssl=?11, pop3login=?12," +
                           "smtpHost=?13, smtpPort=?14, smtpSsl=?15, smtpLogin=?16 WHERE id=?17"

                         : "INSERT INTO accounts (name, email, refresh, deleteMode," +
                           "color, signature, isImap, pop3auth," + 
                           "pop3host, pop3port, pop3ssl, pop3login, " + 
                           "smtpHost, smtpPort, smtpSsl, smtpLogin) " +
                           "VALUES (?1, ?2, ?3, ?4, ?5, ?6, ?7, ?8, ?9, ?10, ?11, ?12, ?13, ?14, ?15, ?16)";

    var statement = connection.createStatement(sql);

    bindString(statement, 0, account.name);
    bindString(statement, 1, account.email);
    statement.bindDoubleParameter(2, account.refresh);
    bindString(statement, 3, account.deleteMode);
    bindString(statement, 4, account.color);
    bindString(statement, 5, account.signature);
    statement.bindInt32Parameter(6, account.pop3.isImap);
    statement.bindInt32Parameter(7, account.pop3.auth);
    bindString(statement, 8, account.pop3.host);
    statement.bindInt32Parameter(9, account.pop3.port);
    statement.bindInt32Parameter(10, account.pop3.ssl);
    bindString(statement, 11, account.pop3.login);
    bindString(statement, 12, account.smtp.host);
    statement.bindInt32Parameter(13, account.smtp.port);
    statement.bindInt32Parameter(14, account.smtp.ssl);
    bindString(statement, 15, account.smtp.login);
    if (account.id) statement.bindInt32Parameter(16, account.id);
    statement.execute();

    if (!account.id) {
      var statement = connection.createStatement("SELECT last_insert_rowid() FROM accounts");
      try {
        if (statement.executeStep()) account.id = statement.getInt32(0);
      }
      finally {
        statement.reset();
      }
    }
    SmPasswordManager.setPassword(account.id + "/pop3/", account.pop3.password);
    SmPasswordManager.setPassword(account.id + "/smtp/", account.smtp.password);

    // Invalidate accounts list cache
    accounts = null;
  }

  // Doesn't delete account messages
  this.deleteAccount = function(account) {
    // Purge deleted messages
    var sql = "DELETE FROM messages WHERE deleted=1 AND accountId=?1";
    var statement = connection.createStatement(sql);
    statement.bindInt32Parameter(0, account.id);
    statement.execute();

    var statement = connection.createStatement("DELETE FROM accounts WHERE id=?1");
    statement.bindInt32Parameter(0, account.id);
    statement.execute();

    SmPasswordManager.removePassword(account.id + "/pop3/");
    SmPasswordManager.removePassword(account.id + "/smtp/");

    if (account.color) {
      SmFile.deleteImage(account.getUnreadIcon());
      SmFile.deleteImage(account.getReadIcon());
    }

    // Invalidate accounts list cache
    accounts = null;
  }

  this.messageExists = function(accountId, uid) {
    var statement = connection.createStatement("SELECT uid FROM messages WHERE accountId=?1 AND uid=?2");
    try {
      statement.bindInt32Parameter(0, accountId);
      bindString(statement, 1, uid);
      return statement.executeStep();
    }
    finally {
      statement.reset();
    }
  }

  // "Search" is performed in all folders.
  // Order by colors uses additional column noColor, so that no-color items appear at bottom

  this.getMessages = function(folderId, search, order, accountId) {
    var sql = "SELECT id, accountId, uid, fromm, too, cc, bcc, subject, date, size," +
              " charset, read, deleted, attachmentsDir, attachmentsCount, folderId," +
              " partial, template, color = 0 AS noColor, color, returnReceiptTo" +
              " FROM messages WHERE deleted = 0 AND " +
              (search ? "(subject LIKE ?1 OR html LIKE ?1)" :
               folderId ? "folderId=?1" : "accountId=?1") +
              " ORDER BY " + (order || "id DESC");

    var statement = connection.createStatement(sql);
    try {
      var messages = new Array();
      if (search) bindString(statement, 0, "%" + search + "%");
      else statement.bindInt32Parameter(0, folderId || accountId);

      while(statement.executeStep()) {
        var id = statement.getInt32(0);
        messages.push(new SmMessage(id,
                                    statement.getInt32(1),
                                    getString(statement, 2),
                                    getString(statement, 3),
                                    getString(statement, 4),
                                    getString(statement, 5),
                                    getString(statement, 6),
                                    getString(statement, 7),
                                    new Date(statement.getInt64(8)),
                                    statement.getDouble(9),
                                    getString(statement, 10),
                                    statement.getInt32(11),
                                    statement.getInt32(12),
                                    getString(statement, 13),
                                    statement.getInt32(14),
                                    null,
                                    statement.getInt32(15),
                                    statement.getInt32(16),
                                    statement.getInt32(17),
                                    statement.getInt32(19),
                                    getString(statement, 20)));
      }
      return messages;
    }
    finally {
      statement.reset();
    }
  }

  this.getAccountMessages = function(accountId) {
    return this.getMessages(null, null, null, accountId);
  }

  this.saveMessage = function(message) {
    // Do not really delete message, keep its uid in DB so that it's not downloaded from server again
    if (message.deleted) {
      SmFile.deleteAttachments(message.attachmentsDir);

      message.from = null;
      message.to = null;
      message.cc = null;
      message.bcc = null;
      message.subject = null;
      message.date = null;
      message.size = null;
      message.charset = null;
      message.read = null;
      message.attachmentsDir = null;
      message.attachmentsCount = null;
      message.html = null;
      message.folderId = null;
      message.partial = null;
      message.template = null;
      message.color = null;
      message.returnReceiptTo = null;
    }
    else SmFile.saveAttachment(message.attachmentsDir,
                               "message.txt",
                               message.source || message.toString());

    var sql = message.id ? "UPDATE messages SET accountId=?1, uid=?2, fromm=?3, too=?4," +
                           "cc=?5, bcc=?6, subject=?7, date=?8, size=?9, charset=?10," +
                           "read=?11, deleted=?12, attachmentsDir=?13, attachmentsCount=?14," +
                           "html=?15, folderId=?16, partial=?17, template=?18, color=?19," +
                           "returnReceiptTo=?20 WHERE id=?21"
                         : "INSERT INTO messages (accountId, uid, fromm, too, cc, bcc," +
                           "subject, date, size, charset, read, deleted, attachmentsDir," +
                           "attachmentsCount, html, folderId, partial, template, color," +
                           "returnReceiptTo) VALUES (?1, ?2, ?3, ?4, ?5, ?6, ?7, ?8, ?9," +
                           "?10, ?11, ?12, ?13, ?14, ?15, ?16, ?17, ?18, ?19, ?20)";

    var statement = connection.createStatement(sql);

    statement.bindInt32Parameter(0, message.accountId);
    bindString(statement, 1, message.uid);
    bindString(statement, 2, message.from);
    bindString(statement, 3, message.to);
    bindString(statement, 4, message.cc);
    bindString(statement, 5, message.bcc);
    bindString(statement, 6, message.subject);
    statement.bindInt64Parameter(7, Date.parse(message.date));
    statement.bindDoubleParameter(8, !message.deleted && message.getSize());
    bindString(statement, 9, message.charset);
    statement.bindInt32Parameter(10, message.read);
    statement.bindInt32Parameter(11, message.deleted);
    bindString(statement, 12, message.attachmentsDir);
    statement.bindInt32Parameter(13, message.attachmentsCount);
    bindString(statement, 14, message.html);
    statement.bindInt32Parameter(15, message.folderId);
    statement.bindInt32Parameter(16, message.partial);
    statement.bindInt32Parameter(17, message.template);
    statement.bindInt32Parameter(18, message.color);
    bindString(statement, 19, message.returnReceiptTo);
    if (message.id) statement.bindInt32Parameter(20, message.id);
    statement.execute();

    if (!message.id) {
      var statement = connection.createStatement("SELECT last_insert_rowid() FROM messages");
      try {
        if (statement.executeStep()) message.id = statement.getInt32(0);
      }
      finally {
        statement.reset();
      }
    }
  }

  this.getMessage = function(id) {
    var sql = "SELECT id, accountId, uid, fromm, too, cc, bcc, subject, date, size, " +
              "charset, read, deleted, attachmentsDir, attachmentsCount, html, folderId," +
              "partial, template, color, returnReceiptTo FROM messages WHERE id=?1";      
    var statement = connection.createStatement(sql);
    try {
      statement.bindInt32Parameter(0, id);
      if (statement.executeStep()) return new SmMessage(id,
                                                        statement.getInt32(1),
                                                        getString(statement, 2),
                                                        getString(statement, 3),
                                                        getString(statement, 4),
                                                        getString(statement, 5),
                                                        getString(statement, 6),
                                                        getString(statement, 7),
                                                        new Date(statement.getInt64(8)),
                                                        statement.getDouble(9),
                                                        getString(statement, 10),
                                                        statement.getInt32(11),
                                                        statement.getInt32(12),
                                                        getString(statement, 13),
                                                        statement.getInt32(14),
                                                        getString(statement, 15),
                                                        statement.getInt32(16),
                                                        statement.getInt32(17),
                                                        statement.getInt32(18),
                                                        statement.getInt32(19),
                                                        getString(statement, 20));
    }
    finally {
      statement.reset();
    }
  }

  this.saveMessageState = function(message) {
    var sql = "UPDATE messages SET read=?1, folderId=?2, color=?3, returnReceiptTo=?4 WHERE id=?5";
    var statement = connection.createStatement(sql);
    statement.bindInt32Parameter(0, message.read);
    statement.bindInt32Parameter(1, message.folderId);
    statement.bindInt32Parameter(2, message.color);
    bindString(statement, 3, message.returnReceiptTo);
    statement.bindInt32Parameter(4, message.id);
    statement.execute();
  }

  this.deleteMessage = function(message) {
    if (message.uid && this.getAccount(message.accountId)) {
      message.deleted = true;
      this.saveMessage(message);
    } else {
      SmFile.deleteAttachments(message.attachmentsDir);
      var statement = connection.createStatement("DELETE FROM messages WHERE id=?1");
      statement.bindInt32Parameter(0, message.id);
      statement.execute();
    }
  }

  this.getFolders = function() {
    var statement = connection.createStatement("SELECT id, parentId, name FROM folders");
    try {
      var folders = new Array();
      while(statement.executeStep()) {
        var parentId = statement.getInt32(1);
        if (!folders[parentId]) {
          folders[parentId] = new Array();
        }
        folders[parentId].push(new SmFolder(statement.getInt32(0),
                                            parentId,
                                            getString(statement, 2)));
      }
      return folders;
    }
    finally {
      statement.reset();
    }
  }

  this.saveFolder = function(folder) {
    var sql = folder.id ? "UPDATE folders SET parentId = ?1, name = ?2 WHERE id=?3"
                        : "INSERT INTO folders (parentId, name) VALUES (?1, ?2)";
    var statement = connection.createStatement(sql);
    statement.bindInt32Parameter(0, folder.parentId);
    bindString(statement, 1, folder.name);
    if (folder.id) statement.bindInt32Parameter(2, folder.id);
    statement.execute();

    if (!folder.id) {
      var statement = connection.createStatement("SELECT last_insert_rowid() FROM folders");
      try {
        if (statement.executeStep()) folder.id = statement.getInt32(0);
      }
      finally {
        statement.reset();
      }
    }
  }

  this.deleteFolder = function(folder, childrenOnly) {
    // Recursively delete child folders
    var children = this.getFolders()[folder.id];
    for(var i in children) {
      this.deleteFolder(children[i]);
    }
    if (!childrenOnly) {
      var statement = connection.createStatement("DELETE FROM folders WHERE id=?1");
      statement.bindInt32Parameter(0, folder.id);
      statement.execute();
    }
  }
}
