var DefaultTreeView = {
  setTree: function(treebox) {
    this.treebox = treebox;
    if (treebox) this.tree = treebox.treeBody.parentNode;
  },
  isContainer: function(row) { return false; },
  isSeparator: function(row) { return false; },
  isSorted: function() { return false; },
  cycleHeader: function(column) {},
  cycleCell: function(row, column) { this.tree.view.selection.select(row); },
  getLevel: function(row) { return 0; },
  getImageSrc: function(row, column) { return null; },
  getRowProperties: function(row, props) {},
  getCellProperties: function(row, column, props) {},
  getColumnProperties: function(column, props) {},
  canDrop: function(row, orientation) { return false; },

  extend: function(treeView) {
    for(var i in this) {
      if (treeView[i] == null) treeView[i] = this[i];
    }
  },

  atomService: Components.classes["@mozilla.org/atom-service;1"]
               .getService(Components.interfaces.nsIAtomService),

  nextDirection: {
    ""           : "ascending",
    "ascending"  : "descending",
    "descending" : ""
  },

  getSortedColumn: function() {
    var column = messageTree.columns.getSortedColumn();
    if (column) {
      return $(column.id);
    }
  },

  forEachSelectedItem: function(onItem) {
    var start = new Object();
    var end = new Object();
    var rangeCount = this.tree.view.selection.getRangeCount();

    this.treebox.beginUpdateBatch();
    for (var i = rangeCount - 1; i >= 0; i--){
      this.tree.view.selection.getRangeAt(i, start, end);
      for (var j = end.value; j >= start.value; j--) {
        onItem(j);
      }
    }
    this.treebox.endUpdateBatch();
  },

  // Event handling

  handleEvent: function(event) {
    if (event.type == "dblclick") {
      var row = {}, column = {}, child = {};
      this.treebox.getCellAt(event.clientX, event.clientY, row, column, child);
      if (row.value != -1) this.action();
    }
    else if (event.type == "keypress") {
      var char = String.fromCharCode(event.charCode);

      if (!event.ctrlKey && !event.shiftKey && !event.altKey) {
        switch (event.keyCode) {
          case event.DOM_VK_RETURN: return this.action();
          case event.DOM_VK_DELETE: return this.delete();
          case event.DOM_VK_INSERT: return this.select();
        }
        switch (char) {
          case "*": return this.selectAll();
        }
      }
      if (event.ctrlKey && !event.shiftKey && !event.altKey) {
        switch (char) {
          case "a": return this.selectAll();
        }
      }
    }
  },

  action: function() {},
  delete: function() {},
  select: function() {
    this.tree.view.selection.rangedSelect(this.tree.currentIndex,
                                          this.tree.currentIndex, true);
    if (this.tree.currentIndex < this.rowCount - 1) {
      this.tree.currentIndex++;
    }
  },
  selectAll: function() {
    this.tree.view.selection.selectAll();
  }
}
