function $(id) {
  return document.getElementById(id);
}

// "Wed Oct 10 2007 14:52:53 GMT+0300" => "Wed, 10 Oct 2007 14:52:53 +0300"
Date.prototype.toMIMEString = function() {
  return this.toString().replace(/(^\w+) (\w+) (\w+) (.*) GMT(.*)/, "$1, $3 $2 $4 $5");
}

var SmUtils = {
  // Make a reversed copy of an array, preserving keys
  reverse: function(array) {
    var reverseKeys = new Array();
    for(var i in array) {
      reverseKeys.unshift(i);
    }
    var result = new Array();
    for(var i in reverseKeys) {
      var key = reverseKeys[i];
      result[key] = array[key];
    }
    return result;
  },

  merge: function(array1, array2) {
    var result = new Array();
    for(var i in array1) {
      result[i] = array1[i];
    }

    for(var i in array2) {
      if (array2[i] != null) result[i] = array2[i];
    }
    return result;
  },

  clearElement: function(element) {
    for(var i = element.childNodes.length - 1; i >= 0; i--) {
      element.removeChild(element.childNodes[i]);
    }
  },

  alert: function(message, title) {
    var prompts = Components.classes["@mozilla.org/embedcomp/prompt-service;1"]
                  .getService(Components.interfaces.nsIPromptService);
    return prompts.alert(window, title || "Simple Mail", message);
  },

  confirm: function(message, wnd) {
    var prompts = Components.classes["@mozilla.org/embedcomp/prompt-service;1"]
                  .getService(Components.interfaces.nsIPromptService);
    return prompts.confirm(wnd || window, "Simple Mail", message);
  },

  confirmCheck: function(message, checkMessage, check, wnd) {
    var prompts = Components.classes["@mozilla.org/embedcomp/prompt-service;1"]
                  .getService(Components.interfaces.nsIPromptService);
    return prompts.confirmCheck(wnd || window, "Simple Mail", message, checkMessage, check);
  },

  prompt: function(message, value) {
    var prompts = Components.classes["@mozilla.org/embedcomp/prompt-service;1"]
                  .getService(Components.interfaces.nsIPromptService);
    var message = (SmBundle.getString(message) || message) + ":";
    return prompts.prompt(window, "Simple Mail", message, value, null, { value: true });
  },

  popup: function(image, title, message, onClick) {
    var listener = { observe: function(subject, topic, data) {
                                if (topic == "alertclickcallback" && onClick) {
                                  onClick();
                                }
                              }};
    var alertsService = Components.classes["@mozilla.org/alerts-service;1"]
                        .getService(Components.interfaces.nsIAlertsService);
    alertsService.showAlertNotification(image, title, message, true, null, listener);
  },

  error: function(message, params) {
    var simpleMailError = SmBundle.getString("simpleMailError");
    var message = SmBundle.getString(message, params);

    if (SmPrefs.getBoolPref("errorsInPopup")) {
      var browserWindow = SmUtils.getThisWindow().browserWindow;
      SmUtils.popup("chrome://simplemail/skin/status/errors_large.png",
                    simpleMailError, message,
                    function() {
                      browserWindow.toJavaScriptConsole();
                    });
    } else {
      SmUtils.alert(message, simpleMailError);
    }

    message = simpleMailError + ": " + message;
    Components.utils.reportError(message);
    return message;
  },

  getThisWindow: function() {
    var browser, tab;
    var wm = Components.classes["@mozilla.org/appshell/window-mediator;1"]
             .getService(Components.interfaces.nsIWindowMediator);
    var browserWindow = wm.getMostRecentWindow("navigator:browser");

    if (browserWindow) {
      browser = browserWindow.getBrowser();

      var tabs = browser.tabContainer.childNodes;
      for(var i = 0; i < tabs.length; i++) {
        if (tabs[i].linkedBrowser.contentDocument == document) tab = tabs[i];
      }
    }
    return { "window"       : window,
             "browserWindow": browserWindow,
             "browser"      : browser,
             "tab"          : tab };
  },

  toUTF8: function(text, encoding) {
    try {
      var utf = Components.classes["@mozilla.org/intl/utf8converterservice;1"]
                .getService(Components.interfaces.nsIUTF8ConverterService);
      return utf.convertStringToUTF8(text, encoding, false);
    }
    catch(e) {
      return text;
    }
  },

  fromUTF8: function(text, encoding) {
    try {
      var converter = Components.classes["@mozilla.org/intl/scriptableunicodeconverter"]
                      .getService(Components.interfaces.nsIScriptableUnicodeConverter);
      converter.charset = encoding;
      var result = converter.ConvertFromUnicode(text);
      return result + converter.Finish();
    }
    catch(e) {
      return text;
    }
  },

  toJSON: function(object, indent, level) {
    if (!level) level = 0;
    if (!indent) indent = " ";
    var json = "";

    for(var i in object) {
      var value = object[i];
      if (i[0] != "_" &&     // Field names beginning with "_" are ignored
          value != null &&
          typeof value != "function")
      {
        if (json) json += "," + indent;
        json += "'" + i + "': ";

        switch(typeof value) {
          case "number":
          case "boolean":
              json += value;
              break;
          case "string":
              json += "'" + value.replace(/([\\\'\n])/g, "\\$1") + "'";
              break;
          default:
              json += this.toJSON(value, indent, level + 1);
              break;
        }
      }
    }
    json = "{" + indent + json + indent + "}";
    return level ? json : "(" + json + ")";
  },

  getSelectedHtml: function() {
    function link(href, title) {
      if (!href) href = content.document.location.href;
      if (!title) title = content.document.title;
      return "<a href='" + href + "'>" + (title || href) + "</a>";
    }
    var html;
    var selection = content.getSelection();

    if (selection.rangeCount) {
      var fragment = selection.getRangeAt(0).cloneContents();
      var div = content.document.createElement("div");
      div.appendChild(fragment);
      html = div.innerHTML;
    }
    if (!html && gContextMenu.target.nodeName == "IMG") {
      html = "<img src='" + gContextMenu.target.src + "'>";
    }
    if (html) {
      return html + "<br><br>" + SmBundle.getString("source") + ": " + link();
    }
    else if (gContextMenu.target.nodeName == "A") {
      return link(gContextMenu.target.href, gContextMenu.target.innerHTML);
    }
    else return link();
  },

  whereToOpenLink: function(event) {
    if (!event) return "";

    if (!event.button) {
      return event.shiftKey ? (event.ctrlKey ? "tabshifted" : "window")
                            : (event.ctrlKey ? "tab" : "");
    } else if (event.button == 1) {
      return event.shiftKey ? "tab" : "tabshifted";
    }
  },

  // Works in both FF & SeaMonkey

  openUILink: function(url, where) {
    if (where instanceof Event) {
      where = SmUtils.whereToOpenLink(where);
    }
    switch(where) {
      case "tabshifted": return gBrowser.addTab(url);
      case "tab":        return gBrowser.selectedTab = gBrowser.addTab(url);
      case "window":     return open(url);
      default:           return loadURI(url);
    }
  },

  timePassed: function(date) {
    if (date) {
      var seconds = Math.floor((new Date() - date) / 1000);
      var hours = Math.floor(seconds / 3600);
      seconds = seconds - hours * 3600;
      var minutes = Math.floor(seconds / 60);
      seconds = seconds - minutes * 60;

      function pad(number) {
        return (number < 10) ? "0" + number : number;
      }

      return pad(hours) + ":" + pad(minutes) + ":" + pad(seconds);
    }
  },

  md5: function(string) {
    // Build array of character codes to MD5
    var array = new Array();
    for(var i = 0; i < string.length; i++) {
      array.push(string.charCodeAt(i));
    } 

    var hash = Components.classes["@mozilla.org/security/hash;1"]
               .createInstance(Components.interfaces.nsICryptoHash);
    hash.init(hash.MD5);
    hash.update(array, array.length);
    var binary = hash.finish(false);

    // Unpack the binary data bin2hex style
    var result = "";
    for (var i = 0; i < binary.length; ++i) {
      var c = binary.charCodeAt(i);
      var ones = c % 16;
      var tens = c >> 4;
      result += String.fromCharCode(tens + (tens > 9 ? 87 : 48)) +
                String.fromCharCode(ones + (ones > 9 ? 87 : 48));
    }
    return result;
  },

  size2string: function(value) {
    return value >= 1000000 ? (value / 1000000).toFixed(1) + " " + SmBundle.getString("mb") :
           value >= 1000 ? (value / 1000).toFixed(1) + " " + SmBundle.getString("kb")
                         : value + " " + SmBundle.getString("bytes");
  },

  // Highlight given phrase, skipping html tags & entities

  highlight: function(phrase, html) {
    var regexp = new RegExp("(<.*?>|&.*?;)|(" + phrase + ")", "gi");
    return html.replace(regexp, function($0, $1, $2) {
      return $1 || "<span class='highlight'>" + $2 + "</span>";
    });
  },

  copyToClipboard: function(string) {
    var clipboard = Components.classes["@mozilla.org/widget/clipboardhelper;1"]
                    .getService(Components.interfaces.nsIClipboardHelper);
    clipboard.copyString(string);
  },

  lastUniqueID: 0,
  uniqueID: function() {
    return SmUtils.lastUniqueID = Math.max(SmUtils.lastUniqueID + 1, new Date().getTime());
  },

  getFontList: function() {
    var enumerator = Components.classes["@mozilla.org/gfx/fontenumerator;1"]
                     .getService(Components.interfaces.nsIFontEnumerator);
    var count = {};
    return enumerator.EnumerateAllFonts(count);
  }
}

var SmBundle = {
  bundle: Components.classes["@mozilla.org/intl/stringbundle;1"]
          .getService(Components.interfaces.nsIStringBundleService)
          .createBundle("chrome://simplemail/locale/simplemail.properties"),

  getString: function(name, params) {
    try {
      return params ? SmBundle.bundle.formatStringFromName(name, params, params.length)
                    : SmBundle.bundle.GetStringFromName(name);
    }
    catch(e) {
      return null;
    }
  }
}
